import os
import runpy
import sys
from pathlib import Path

deps_dir = Path(__file__).parent / "deps"
if not deps_dir.exists():
    deps_dir.mkdir()

if not (deps_dir / "venvstarter.py").exists():
    if "PIP_REQUIRE_VIRTUALENV" in os.environ:
        del os.environ["PIP_REQUIRE_VIRTUALENV"]
    os.system(f"{sys.executable} -m pip install venvstarter -t {deps_dir}")

venvstarter_module = runpy.run_path(str(deps_dir / "venvstarter.py"))

wanted_version = "0.12.0"

upgrade = False
VERSION = venvstarter_module.get("VERSION")
if VERSION is None:
    upgrade = True
else:
    Version = venvstarter_module["Version"]
    if Version(VERSION) != Version(wanted_version):
        upgrade = True

if upgrade:
    os.system(f"{sys.executable} -m pip install -U 'venvstarter=={wanted_version}' -t {deps_dir}")

manager = runpy.run_path(str(deps_dir / "venvstarter.py"))["manager"]



import os
import re
import sys
import tempfile
from pathlib import Path

load_grammar_line = '_GRAMMAR_FILE = os.path.join(os.path.dirname(__file__), "Grammar.txt")'


def is_supported_black_version():
    # I'm caring too much about internals to be confident
    # about future compatibility
    try:
        import black  # noqa
        import blib2to3  # noqa
        import importlib_resources  # noqa
    except ImportError:
        return

    return not black.COMPILED and black.__version__ == "22.10.0"


def maybe_modify_black():
    if os.environ.get("NOSE_OF_YETI_BLACK_COMPAT") == "true":
        modify_black()


def replace_pygram():
    from importlib.machinery import ModuleSpec, SourceFileLoader
    from importlib.util import module_from_spec

    import importlib_resources as resources

    for k, m in list(sys.modules.items()):
        if "blib2to3" in k or "black" in k:
            del sys.modules[k]

    with resources.as_file(
        resources.files("noseOfYeti").joinpath("black", "Grammar.noy.txt")
    ) as noy_grammar_path:
        if noy_grammar_path.exists():
            location = Path(__import__("blib2to3").__file__).parent / "Grammar.noy.txt"
            current = None
            if location.exists():
                with open(location) as crnt:
                    current = crnt.read()

            with open(noy_grammar_path) as ngp:
                content = ngp.read()
                if content != current:
                    with open(location, "w") as fle:
                        fle.write(content)

        with tempfile.TemporaryDirectory() as directory:
            location = Path(directory) / "pygram.py"
            with open(location, "w") as fle:
                noy_grammar_line = (
                    '_GRAMMAR_FILE = os.path.join(os.path.dirname(__file__), "Grammar.noy.txt")'
                )
                fle.write(
                    (resources.files("blib2to3") / "pygram.py")
                    .read_text()
                    .replace(load_grammar_line, noy_grammar_line)
                )

            loader = SourceFileLoader("pygram", location)
            mod = module_from_spec(ModuleSpec("pygram", loader))
            mod.__file__ = str(resources.files("blib2to3").joinpath("pygram.py"))
            mod.__package__ = "blib2to3"
            loader.exec_module(mod)
            sys.modules["blib2to3.pygram"] = mod


def modify_black(spec_codec=None):
    """
    This will make it so calling black after this in the same session will
    deal with noseOfYeti spec files.
    """
    if not is_supported_black_version():
        return

    import blib2to3
    from blib2to3 import pygram

    if not str(pygram.__loader__.path).startswith(str(Path(blib2to3.__file__).parent)):
        return
    del pygram
    del blib2to3

    if spec_codec is None:
        from noseOfYeti.tokeniser.spec_codec import codec, register

        spec_codec = codec()
        register(transform=False)

    replace_pygram()

    import black
    from black.linegen import LineGenerator
    from blib2to3.pgen2 import parse as blibparse

    token = blibparse.token
    Parser = blibparse.Parser

    class ModifiedLineGenerator(LineGenerator):
        def visit_setup_teardown_stmts(self, node):
            yield from self.line()
            yield from self.visit_default(node)

        def visit_describe_stmt(self, node):
            yield from self.line()
            yield from self.visit_default(node)

        def visit_it_stmt(self, node):
            yield from self.line()
            yield from self.visit_default(node)

    class ModifiedParser(Parser):
        def classify(self, type, value, context):
            special = [
                "it",
                "ignore",
                "context",
                "describe",
                "before_each",
                "after_each",
            ]
            if type == token.NAME and value in special:
                dfa, state, node = self.stack[-1]
                if node and node[-1]:
                    if node[-1][-1].type < 256 and node[-1][-1].type not in (
                        token.INDENT,
                        token.DEDENT,
                        token.NEWLINE,
                        token.ASYNC,
                    ):
                        return [self.grammar.tokens.get(token.NAME)]
            return super().classify(type, value, context)

    original_assert_equivalent = black.assert_equivalent

    def assert_equivalent(src, dst):
        if src and re.match(r"#\s*coding\s*:\s*spec", src.split("\n")[0]):
            src = spec_codec.translate(src, transform=True)
            dst = spec_codec.translate(dst, transform=True)
        original_assert_equivalent(src, dst)

    blibparse.Parser = ModifiedParser
    black.LineGenerator = ModifiedLineGenerator
    black.assert_equivalent = assert_equivalent



# coding: spec

import unittest

import pytest


@pytest.fixture()
def hi():
    return "hi"


it "one", hi:
    assert hi == "hi"

describe "two":
    it "three", hi:
        assert hi == "hi"

    it "four", hi:
        assert hi == "hi"

    describe "five":
        it "six", hi:
            assert hi == "hi"

    describe "seven":
        __only_run_tests_in_children__ = True

        it "eight", hi, expected:
            assert hi == expected

        it "nine", hi, expected:
            assert hi == expected

        describe "ten":

            @pytest.fixture()
            def hi(self):
                return "hello"

            @pytest.fixture()
            def expected(self):
                return "hello"

        describe "eleven":

            @pytest.fixture()
            def hi(self):
                return "yo"

            @pytest.fixture()
            def expected(self):
                return "yo"

describe unittest.TestCase, "twelve":
    it "thirteen":
        assert True



from noseOfYeti.tokeniser.spec_codec import TokeniserCodec, register
from noseOfYeti.tokeniser.support import TestSetup
from noseOfYeti.tokeniser.tokeniser import Tokeniser

__all__ = [
    "register",
    "Tokeniser",
    "TestSetup",
    "TokeniserCodec",
]



import re

import pytest


class TestMismatchedBrackets:
    def test_it_knows_about_mismatched_square_from_parenthesis(self):
        original = """
        def wat(self:
            pass

        things = "]"

        def things]self:
            pass
        """

        expected = re.escape(
            "Trying to close the wrong type of bracket. Found ']' (line 6, column 10) instead of closing a '(' (line 1, column 7)"
        )

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_mismatched_square_from_curly(self):
        original = """
        def wat{self:
            pass

        def things]self:
            pass
        """

        expected = re.escape(
            "Trying to close the wrong type of bracket. Found ']' (line 4, column 10) instead of closing a '{' (line 1, column 7)"
        )

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_mismatched_parenthesis_from_square(self):
        original = """
        def wat[self:
            pass

        def things)self:
            pass
        """

        expected = re.escape(
            "Trying to close the wrong type of bracket. Found ')' (line 4, column 10) instead of closing a '[' (line 1, column 7)"
        )

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_hanging_square(self):
        original = """
        def wat(self):
            pass

        def things]self:
            pass
        """

        expected = re.escape("Found a hanging ']' on line 4, column 10")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_hanging_parenthesis(self):
        original = """
        def wat(self)):
            pass
        """

        expected = re.escape("Found a hanging ')' on line 1, column 13")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_hanging_curly(self):
        original = """
        class Wat:
            def __init__(self):
                self.d = {1: 2}}
        """

        expected = re.escape("Found a hanging '}' on line 3, column 23")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_unclosed_parenthesis(self):
        original = """
        def thing(self):
            pass

        def wat(self:
            pass
        """

        expected = re.escape("Found an open '(' (line 4, column 7) that wasn't closed")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_unclosed_square(self):
        original = """
        def thing(self):
            pass

        things = [1, 2
        """

        expected = re.escape("Found an open '[' (line 4, column 9) that wasn't closed")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")

    def test_it_knows_about_unclosed_curly(self):
        original = """
        def thing(self):
            pass

        things = [1, 2]

        stuff = {1: 2
        """

        expected = re.escape("Found an open '{' (line 6, column 8) that wasn't closed")

        with pytest.raises(SyntaxError, match=expected):
            pytest.helpers.assert_conversion(original, "")



import codecs
import encodings
import re
import sys
import traceback
from encodings import utf_8
from io import StringIO
from tokenize import untokenize

from noseOfYeti.tokeniser.tokeniser import Tokeniser

regexes = {
    "whitespace": re.compile(r"\s*"),
    "only_whitespace": re.compile(r"^\s*$"),
    "encoding_matcher": re.compile(r"#\s*coding\s*:\s*spec"),
    "leading_whitespace": re.compile(r"^(\s*)[^\s]"),
}


class TokeniserCodec:
    """Class to register the spec codec"""

    def __init__(self, tokeniser):
        self.tokeniser = tokeniser

        self.transform = True
        self.codec = self.get_codec()

    def translate(self, value, transform=None):
        if isinstance(value, str):
            value = value.encode()

        return self.codec.decode(
            value, return_tuple=False, transform=self.transform if transform is None else transform
        )

    def register(self):
        def search_function(s):
            """Determine if a file is of spec encoding and return special CodecInfo if it is"""
            if s != "spec":
                return None
            return self.codec

        # Do the register
        codecs.register(search_function)

    def get_codec(self):
        """Register spec codec"""
        # Assume utf8 encoding
        utf8 = encodings.search_function("utf8")

        class StreamReader(utf_8.StreamReader):
            """Used by cPython to deal with a spec file"""

            def __init__(sr, stream, *args, **kwargs):
                codecs.StreamReader.__init__(sr, stream, *args, **kwargs)
                if self.transform:
                    data = self.dealwith(sr.stream.readline)
                    sr.stream = StringIO(data)

        def _decode(text, *args, transform=None, **kwargs):
            transform = self.transform if transform is None else transform

            if not transform:
                return utf8.decode(text, *args, **kwargs)

            if hasattr(text, "tobytes"):
                text = text.tobytes().decode()
            else:
                text = text.decode()

            reader = StringIO(text)

            # Determine if we need to have imports for this string
            # It may be a fragment of the file
            has_spec = regexes["encoding_matcher"].search(reader.readline())
            no_imports = not has_spec
            reader.seek(0)

            data = self.dealwith(reader.readline, no_imports=no_imports)

            # If nothing was changed, then we want to use the original file/line
            # Also have to replace indentation of original line with indentation of new line
            # To take into account nested describes
            if text and not regexes["only_whitespace"].match(text):
                if regexes["whitespace"].sub("", text) == regexes["whitespace"].sub("", data):
                    bad_indentation = regexes["leading_whitespace"].search(text).groups()[0]
                    good_indentation = regexes["leading_whitespace"].search(data).groups()[0]
                    data = f"{good_indentation}{text[len(bad_indentation) :]}"

            # If text is empty and data isn't, then we should return text
            if len(text) == 0 and len(data) == 1:
                return "", 0

            # Return translated version and it's length
            return data, len(data)

        def decode(text, *args, return_tuple=True, transform=None, **kwargs):
            ret = _decode(text, *args, transform=transform, **kwargs)
            if return_tuple:
                return ret
            else:
                return ret[0]

        class incrementaldecoder(utf8.incrementaldecoder):
            def decode(s, obj, final, **kwargs):
                if not self.transform:
                    return super().decode(obj, final, **kwargs)

                lines = obj.split("\n".encode("utf-8"))
                if re.match(r"#\s*coding:\s*spec", lines[0].decode("utf-8", "replace")) and final:
                    kwargs["return_tuple"] = False
                    return decode(obj, final, **kwargs)
                else:
                    return super().decode(obj, final, **kwargs)

        return codecs.CodecInfo(
            name="spec",
            encode=utf8.encode,
            decode=decode,
            streamreader=StreamReader,
            streamwriter=utf8.streamwriter,
            incrementalencoder=utf8.incrementalencoder,
            incrementaldecoder=incrementaldecoder,
        )

    def dealwith(self, readline, **kwargs):
        """
        Replace the contents of spec file with the translated version
        readline should be a callable object,
        which provides the same interface as the readline() method of built-in file objects
        """
        data = []
        try:
            # We pass in the data variable as an argument so that we
            # get partial output even in the case of an exception.
            self.tokeniser.translate(readline, data, **kwargs)
        except:
            lines = ['msg = r"""']
            for line in traceback.format_exception(*sys.exc_info()):
                lines.append(line.strip())
            lines.append('"""')
            lines.append(r'raise Exception(f"--- internal spec codec error --- \n{msg}")')
            data = "\n".join(lines)
        else:
            # At this point, data is a list of tokens
            data = untokenize(data)

        # python3.9 requires a newline at the end
        data += "\n"

        return data

    def output_for_debugging(self, stream, data):
        """It will write the translated version of the file"""
        with open(f"{stream.name}.spec.out", "w") as f:
            f.write(str(data))


########################
###   CODEC REGISTER
########################

_spec_codec = None


def codec():
    """Return the codec used to translate a file"""
    global _spec_codec
    if _spec_codec is None:
        _spec_codec = TokeniserCodec(Tokeniser())
    return _spec_codec


def register(transform=True):
    """Get a codec and register it in python"""
    do_register = False
    try:
        codecs.lookup("spec")
    except LookupError:
        do_register = True

    cdc = codec()
    cdc.transform = transform

    if do_register:
        cdc.register()



import os
import re
import shutil
import tempfile
from contextlib import contextmanager
from io import StringIO
from textwrap import dedent
from tokenize import untokenize

import pytest

from noseOfYeti.tokeniser.tokeniser import WITH_IT_RETURN_TYPE_ENV_NAME


@pytest.fixture(autouse=True)
def remove_with_it_return_type_env(monkeypatch):
    monkeypatch.delenv(WITH_IT_RETURN_TYPE_ENV_NAME, raising=False)


@pytest.fixture()
def a_temp_file():
    @contextmanager
    def a_temp_file(contents):
        tempfle = None
        try:
            tempfle = tempfile.NamedTemporaryFile(delete=False)
            with open(tempfle.name, "w") as fle:
                fle.write(dedent(contents))

            yield tempfle.name
        finally:
            if tempfle:
                if os.path.exists(tempfle.name):
                    os.remove(tempfle.name)

    return a_temp_file


@pytest.fixture()
def a_temp_dir():
    @contextmanager
    def a_temp_dir():
        tempdir = None
        try:
            tempdir = tempfile.mkdtemp()
            yield tempdir
        finally:
            if tempdir:
                if os.path.exists(tempdir):
                    shutil.rmtree(tempdir)

    return a_temp_dir


def pytest_configure():
    @pytest.helpers.register
    def assert_regex_lines(got_lines, want_lines, lstrip=True, rstrip=False):
        __tracebackhide__ = True

        got_lines = dedent(got_lines)
        want_lines = dedent(want_lines)

        if lstrip:
            got_lines = got_lines.lstrip()
            want_lines = want_lines.lstrip()

        if lstrip:
            got_lines = got_lines.rstrip()
            want_lines = want_lines.rstrip()

        print("GOT LINES\n=========")
        print(got_lines)

        print("\n\n\nWANT LINES\n==========")
        print(want_lines)

        for i, (wl, gl) in enumerate(zip(want_lines.split("\n"), got_lines.split("\n"))):
            try:
                m = re.match(wl, gl)
            except re.error as error:
                pytest.fail(f"Failed to turn line into a regex\nCONVERTING: {wl}\nERROR: {error}")

            if not m:
                pytest.fail(f"line {i} does not match. Wanted:\n{wl}\n\nGot:\n{gl}")

    @pytest.helpers.register
    def assert_lines(got_lines, want_lines, lstrip=True, rstrip=False, rstrip_got_lines=True):
        __tracebackhide__ = True

        got_lines = dedent(got_lines)
        want_lines = dedent(want_lines)

        if lstrip:
            got_lines = got_lines.lstrip()
            want_lines = want_lines.lstrip()

        if lstrip:
            got_lines = got_lines.rstrip()
            want_lines = want_lines.rstrip()

        print("GOT LINES\n=========")
        print(got_lines)

        print("\n\n\nWANT LINES\n==========")
        print(want_lines)

        for i, (wl, gl) in enumerate(zip(want_lines.split("\n"), got_lines.split("\n"))):
            if rstrip_got_lines:
                gl = gl.rstrip()
            if wl != gl:
                pytest.fail(f"line {i} does not match. Wanted:\n{wl}\n\nGot:\n{gl}")

    @pytest.helpers.register
    def assert_conversion(
        original, want_lines, *, tokeniser=None, regex=False, lstrip=True, rstrip=False
    ):
        __tracebackhide__ = True

        from noseOfYeti.tokeniser import Tokeniser

        original = dedent(original)
        if lstrip:
            original = original.lstrip()
        if rstrip:
            original = original.rstrip()

        if tokeniser is None:
            tokeniser = {}

        for give_return_types, ret in ((False, ""), (True, "->None ")):
            orig = original.replace("$RET", ret)
            want = want_lines.replace("$RET", ret)

            if give_return_types:
                kwargs = {**tokeniser, "with_it_return_type": True}
            else:
                kwargs = {**tokeniser, "with_it_return_type": False}

            s = StringIO(orig)
            tok = Tokeniser(**kwargs)
            got_lines = untokenize(tok.translate(s.readline))

            if regex:
                pytest.helpers.assert_regex_lines(got_lines, want, lstrip=lstrip, rstrip=rstrip)
            else:
                pytest.helpers.assert_lines(got_lines, want, lstrip=lstrip, rstrip=rstrip)

    @pytest.helpers.register
    def assert_example(example, convert_to_tabs=False, **kwargs):
        __tracebackhide__ = True

        original, desired = example
        if convert_to_tabs:
            original = original.replace("    ", "\t")
            desired = desired.replace("    ", "\t")

        pytest.helpers.assert_conversion(original, desired, tokeniser=kwargs)



"""
Modified pytest discovery classes.

This is so you can use noseOfYeti to have nested describes without running the
same tests multiple times.
"""
import inspect
from unittest import mock

import pytest
from _pytest.unittest import UnitTestCase

from noseOfYeti.tokeniser.spec_codec import register

pytest_has_instance_class = int(pytest.__version__.split(".")[0]) < 7


def pytest_configure():
    register(transform=True)


def change_item(res, obj):
    original_item_collect = res.collect

    if isinstance(res, UnitTestCase):

        def collect():
            yield from filter_collection(original_item_collect(), obj)

    else:

        def collect():
            res.session._fixturemanager.parsefactories(res)
            yield from filter_collection(original_item_collect(), obj)

    mock.patch.object(res, "collect", collect).start()


@pytest.hookimpl(hookwrapper=True)
def pytest_pycollect_makeitem(collector, name, obj):
    """Make sure we can have nested noseOfYeti describes"""
    outcome = yield
    res = outcome.get_result()

    if not isinstance(res, pytest.Class):
        return

    change_item(res, obj)


def filter_collection(collected, obj):
    for thing in collected:
        if pytest_has_instance_class and isinstance(thing, pytest.Instance):
            if getattr(thing.obj, "is_noy_spec", False):
                change_item(thing, thing.obj)

        if not isinstance(thing, pytest.Function):
            yield thing
            continue

        try:
            isclass = inspect.isclass(obj)
            if not isclass:
                obj = obj.__class__
        except Exception:
            pass

        if obj.__dict__.get("__only_run_tests_in_children__"):
            # Only run these tests in the children, not in this class itself
            continue

        if thing.obj.__name__ in obj.__dict__:
            yield thing
        else:
            method_passed_down = any(
                thing.obj.__name__ in superkls.__dict__
                and getattr(superkls, "__only_run_tests_in_children__", False)
                for superkls in obj.__bases__
            )
            if method_passed_down:
                yield thing



"""
A helper module to access the superclass'
setUp() and tearDown() methods of generated
test classes.
"""

import asyncio


class TestSetup:
    def __init__(self, sup):
        self.sup = sup

    @property
    def setup(self):
        if hasattr(self.sup, "setup"):
            return self.sup.setup
        if hasattr(self.sup, "setUp"):
            return self.sup.setUp

    @property
    def teardown(self):
        if hasattr(self.sup, "teardown"):
            return self.sup.teardown
        if hasattr(self.sup, "tearDown"):
            return self.sup.tearDown

    def sync_before_each(self):
        setup = self.setup
        if setup:
            return setup()

    def sync_after_each(self):
        teardown = self.teardown
        if teardown:
            return teardown()

    async def async_before_each(self):
        res = self.sync_before_each()
        if res and asyncio.iscoroutine(res):
            return await res
        else:
            return res

    async def async_after_each(self):
        res = self.sync_after_each()
        if res and asyncio.iscoroutine(res):
            return await res
        else:
            return res



from tokenize import COMMENT, INDENT, NAME, NEWLINE, OP, STRING, generate_tokens

########################
###   TOKENS IN GENERATOR
########################


def tokens_in(s, strip_it=True):
    closure = {"processed": False}

    def get():
        if closure["processed"]:
            return ""

        closure["processed"] = True
        if strip_it:
            return s.strip()
        else:
            return s

    res = [(t, v) for t, v, _, _, _ in generate_tokens(get)][:-1]
    while res and res[-1][0] == 4:
        res.pop()
    return res


########################
###   TOKENS
########################


class Tokens:
    def __init__(self):
        self.equivalence = {"after_each": "tearDown", "before_each": "setUp"}

        self.before_each = [
            (NAME, "def"),
            (NAME, self.equivalence["before_each"]),
            (OP, "("),
            (NAME, "self"),
            (OP, ")"),
        ]

        self.after_each = [
            (NAME, "def"),
            (NAME, self.equivalence["after_each"]),
            (OP, "("),
            (NAME, "self"),
            (OP, ")"),
        ]

        self.single_colon = (OP, ":")

    ########################
    ###   MAKERS
    ########################

    def make_single(self, name, args, comments, return_type):
        if return_type is True:
            return_type = [(STRING, "->"), (NAME, "None")]
        elif return_type is None:
            return_type = []

        lst = [(NAME, "def"), (NAME, name), (OP, "(")] + [
            (t, n) for t, n, *_ in args if t is not None
        ]

        has_end = True
        has_pass = False

        if lst[-1][1] == ":":
            lst.pop()
        elif return_type and return_type[-1][1] == ":":
            return_type.pop()
        elif lst[-1][1] == "pass" and lst[-2][1] == ":":
            has_pass = True
            lst.pop()
            lst.pop()
        elif return_type and return_type[-1][1] == "pass" and return_type[-2][1] == ":":
            has_pass = True
            return_type.pop()
            return_type.pop()
        else:
            has_end = False

        if not has_end:
            srow = args[-1][-2]
            scol = args[-1][-1]
            raise SyntaxError(f"Found a missing ':' on line {srow}, column {scol}")

        lst.append((OP, ")"))
        lst.extend(return_type)
        lst.append((OP, ":"))

        if has_pass:
            lst.append((NAME, "pass"))

        lst.extend(comments)
        return lst

    def make_describe(self, kls, name):
        lst = [(NAME, "class"), (NAME, name)]
        if kls:
            lst.append((OP, "("))
            lst.extend(tokens_in(kls))
            lst.append((OP, ")"))

        lst.append((OP, ":"))

        return lst

    def make_super(self, indent, kls, method, with_async=False):
        if kls:
            kls = tokens_in(kls)

        prefix = "async" if with_async else "sync"

        result = [(OP, ":"), (NEWLINE, "\n"), (INDENT, indent)]

        if with_async:
            result.append((NAME, "await"))

        result.extend(
            [
                (NAME, "__import__"),
                (OP, "("),
                (OP, '"'),
                (STRING, "noseOfYeti"),
                (OP, '"'),
                (OP, ")"),
                (OP, "."),
                (NAME, "tokeniser"),
                (OP, "."),
                (NAME, "TestSetup"),
                (OP, "("),
                (NAME, "super"),
                (OP, "("),
                (OP, ")"),
                (OP, ")"),
                (OP, "."),
                (NAME, f"{prefix}_{method}"),
                (OP, "("),
                (OP, ")"),
            ]
        )

        return result

    def make_describe_attr(self, describe):
        return [
            (NEWLINE, "\n"),
            (NAME, describe),
            (OP, "."),
            (NAME, "is_noy_spec"),
            (OP, "="),
            (NAME, "True"),
            (COMMENT, " # type: ignore"),
        ]

    def make_name_modifier(self, cleaned, english):
        result = [(NEWLINE, "\n")]

        parts = cleaned.split(".")
        result.append((NAME, parts[0]))

        for part in parts[1:]:
            result.extend([(OP, "."), (NAME, part)])

        result.extend(
            [
                (OP, "."),
                (NAME, "__testname__"),
                (OP, "="),
                (STRING, english),
                (COMMENT, "  # type: ignore"),
            ]
        )

        return result



import pytest

from noseOfYeti.tokeniser.tokeniser import WITH_IT_RETURN_TYPE_ENV_NAME, Tokeniser


class Examples:
    small_example = [
        """
    describe "This":
        before_each:
            self.x = 5
        describe "That":
            before_each:
                self.y = 6
            describe "Meh":
                after_each:
                    self.y = None
        describe "Blah":pass
        describe "async":
            async before_each:
                pass
            async after_each:
                pass
    describe "Another":
        before_each:
            self.z = 8 """,
        """
    class TestThis :
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .x =5
    class TestThis_That (TestThis ):
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .y =6
    class TestThis_That_Meh (TestThis_That ):
        def tearDown (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ();self .y =None
    class TestThis_Blah (TestThis ):pass
    class TestThis_Async (TestThis ):
        async def setUp (self ):
            await __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).async_before_each ();pass
        async def tearDown (self ):
            await __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).async_after_each ();pass
    class TestAnother :
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .z =8

    TestThis .is_noy_spec =True  # type: ignore
    TestThis_That .is_noy_spec =True  # type: ignore
    TestThis_That_Meh .is_noy_spec =True  # type: ignore
    TestThis_Blah .is_noy_spec =True  # type: ignore
    TestThis_Async .is_noy_spec =True  # type: ignore
    TestAnother .is_noy_spec =True  # type: ignore
    """,
    ]

    big_example = [
        """
    describe "This":
        before_each:
            self.x = 5
        it 'should':
            if x:
                pass
            else:
                x += 9
        async it 'supports async its':
            pass
        describe "That":
            before_each:
                self.y = 6
            describe "Meh":
                after_each:
                    self.y = None
                it "should set __testname__ for non alpha names ' $^":
                    pass
                it 'should':
                    if y:
                        pass
                    else:
                        pass
                it 'should have args', arg1, arg2:
                    blah |should| be_good()
        describe "Blah":pass
    ignore "root level $pecial-method*+":
        pass
    describe "Another":
        before_each:
            self.z = 8
        it 'should':
            if z:
                if u:
                    print "hello \
                        there"
                else:
                    print "no"
            else:
                pass
    async it 'supports level 0 async its':
        pass
    """,
        """
    class TestThis :
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .x =5
        def test_should (self )$RET:
            if x :
                pass
            else :
                x +=9
        async def test_supports_async_its (self )$RET:
            pass
    class TestThis_That (TestThis ):
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .y =6
    class TestThis_That_Meh (TestThis_That ):
        def tearDown (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ();self .y =None
        def test_should_set_testname_for_non_alpha_names (self )$RET:
            pass
        def test_should (self )$RET:
            if y :
                pass
            else :
                pass
        def test_should_have_args (self ,arg1 ,arg2 )$RET:
            blah |should |be_good ()
    class TestThis_Blah (TestThis ):pass
    def ignore__root_level_pecial_method ()$RET:
        pass
    class TestAnother :
        def setUp (self ):
            __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .z =8
        def test_should (self )$RET:
            if z :
                if u :
                    print "hello \
                        there"
                else :
                    print "no"
            else :
                pass
    async def test_supports_level_0_async_its ()$RET:
        pass

    TestThis .is_noy_spec =True  # type: ignore
    TestThis_That .is_noy_spec =True  # type: ignore
    TestThis_That_Meh .is_noy_spec =True  # type: ignore
    TestThis_Blah .is_noy_spec =True  # type: ignore
    TestAnother .is_noy_spec =True  # type: ignore

    ignore__root_level_pecial_method .__testname__ ="root level $pecial-method*+"  # type: ignore
    TestThis_That_Meh .test_should_set_testname_for_non_alpha_names .__testname__ ="should set __testname__ for non alpha names ' $^"  # type: ignore
    """,
    ]

    comment_example = [
        """
    assertTileHues(
        self, tiles[0],
        25.0,  25.0,  25.0,  25.0,  25.0,  25.0, # noqa
        18.75, 18.75, 18.75, 18.75, 18.75, 18.75, # noqa
    )

    it "things":
        assertTileHues(
            self, tiles[1],
            25.0,  25.0,  25.0,  25.0,  25.0,  25.0, # noqa
            18.75, 18.75, 18.75, 18.75, 18.75, 18.75, # noqa
        )

        expected = {
            # something
            ("D2", "<d"): lambda s: ("D2", "<d", s)
            # something else
            ,
            ("B2", None): ("B2", None, None),
        }

        def t(n, f, c):
            return expected[(n, f, c)]

        assert True # type: ignore[valid-type]
        assert True

        # one two three
        assert True
    """,
        """
    assertTileHues (
    self ,tiles [0 ],
    25.0 ,25.0 ,25.0 ,25.0 ,25.0 ,25.0 ,# noqa
    18.75 ,18.75 ,18.75 ,18.75 ,18.75 ,18.75 ,# noqa
    )

    def test_things ()$RET:
        assertTileHues (
        self ,tiles [1 ],
        25.0 ,25.0 ,25.0 ,25.0 ,25.0 ,25.0 ,# noqa
        18.75 ,18.75 ,18.75 ,18.75 ,18.75 ,18.75 ,# noqa
        )

        expected ={
        # something
        ("D2","<d"):lambda s :("D2","<d",s )
        # something else
        ,
        ("B2",None ):("B2",None ,None ),
        }

        def t (n ,f ,c ):
            return expected [(n ,f ,c )]

        assert True # type: ignore[valid-type]
        assert True

        # one two three
        assert True
    """,
    ]


class Test_Tokeniser:
    def test_sets_with_it_return_type_appropriately(self, monkeypatch):
        monkeypatch.delenv(WITH_IT_RETURN_TYPE_ENV_NAME, raising=False)
        assert not Tokeniser().with_it_return_type

        monkeypatch.setenv(WITH_IT_RETURN_TYPE_ENV_NAME, "true")
        assert Tokeniser().with_it_return_type

        monkeypatch.setenv(WITH_IT_RETURN_TYPE_ENV_NAME, "false")
        assert not Tokeniser().with_it_return_type
        assert not Tokeniser().with_it_return_type

        monkeypatch.setenv(WITH_IT_RETURN_TYPE_ENV_NAME, "1")
        assert Tokeniser().with_it_return_type
        assert Tokeniser().with_it_return_type

        monkeypatch.setenv(WITH_IT_RETURN_TYPE_ENV_NAME, "0")
        assert not Tokeniser().with_it_return_type
        assert not Tokeniser().with_it_return_type

    def test_gives_describes_noy_specific_attributes(self):
        pytest.helpers.assert_example(
            [
                'describe "Something testable"',
                """
            class TestSomethingTestable :pass

            TestSomethingTestable .is_noy_spec =True  # type: ignore
            """,
            ]
        )


class Test_Tokeniser_Complex:
    def test_works_with_space(self):
        pytest.helpers.assert_example(Examples.small_example)

    def test_works_with_tabs(self):
        pytest.helpers.assert_example(Examples.small_example, convert_to_tabs=True)

    def test_keeps_good_indentation_in_body_with_spaces(self):
        pytest.helpers.assert_example(Examples.big_example)

    def test_keeps_good_indentation_in_body_with_tabs(self):
        pytest.helpers.assert_example(Examples.big_example, convert_to_tabs=True)

    def test_keeps_correct_indentation_with_comments(self):
        pytest.helpers.assert_example(Examples.comment_example, convert_to_tabs=True)



import inspect
import os
import platform
import shlex
import sys
from collections.abc import Callable
from pathlib import Path

here = Path(__file__).parent

if platform.system() == "Windows":
    import mslex  # type:ignore[import]

    shlex = mslex  # noqa


class Command:
    def __call__(self, bin_dir, args) -> None:
        ...


def command(func) -> Callable:
    func.__is_command__ = True
    return func


def run(*args) -> None:
    cmd = " ".join(shlex.quote(str(part)) for part in args)
    print(f"Running '{cmd}'")
    ret = os.system(cmd)
    if ret != 0:
        sys.exit(1)


class App:
    def __init__(self):
        self.commands = {}

        compare = inspect.signature(type("C", (Command,), {})().__call__)

        for name in dir(self):
            val = getattr(self, name)
            if getattr(val, "__is_command__", False):
                assert (
                    inspect.signature(val) == compare
                ), f"Expected '{name}' to have correct signature, have {inspect.signature(val)} instead of {compare}"
                self.commands[name] = val

    def __call__(self, args) -> None:
        bin_dir = Path(sys.executable).parent

        if args and args[0] in self.commands:
            os.chdir(here.parent)
            self.commands[args[0]](bin_dir, args[1:])
            return

        sys.exit(f"Unknown command:\nAvailable: {sorted(self.commands)}\nWanted: {args}")

    @command
    def format(self, bin_dir, args) -> None:
        if not args:
            args = [".", *args]
        run(bin_dir / "black", *args)
        run(bin_dir / "isort", *args)

    @command
    def lint(self, bin_dir, args) -> None:
        run(bin_dir / "pylama", *args)

    @command
    def tests(self, bin_dir, args) -> None:
        if "-q" not in args:
            args = ["-q", *args]

        env = os.environ
        env["NOSE_OF_YETI_BLACK_COMPAT"] = "false"

        files = []
        if "TESTS_CHDIR" in env:
            ags = []
            test_dir = Path(env["TESTS_CHDIR"]).absolute()
            for a in args:
                test_name = ""
                if "::" in a:
                    filename, test_name = a.split("::", 1)
                else:
                    filename = a
                try:
                    p = Path(filename).absolute()
                except:
                    ags.append(a)
                else:
                    if p.exists():
                        rel = p.relative_to(test_dir)
                        if test_name:
                            files.append(f"{rel}::{test_name}")
                        else:
                            files.append(str(rel))
                    else:
                        ags.append(a)
            args = ags
            os.chdir(test_dir)

        run(bin_dir / "pytest", *files, *args)


app = App()

if __name__ == "__main__":
    app(sys.argv[1:])



# coding: spec

# You ensure the file has the coding: spec comment as the first line as above
# and that nose-of-yeti has registered the spec codec

# The codec will then turn what you have written into python code that can be
# executed.

# The test can then be specified using describes and its

from unittest import TestCase


def test_is_possible_to_add_numbers ():
    assert 1 +1 ==2

def test_is_possible_to_add_the_number_three (three =3 ):
    # Contrived example of default arguments
    assert 1 +three ==4

class TestPythonMathematics :
    # ^^ is replaced with "class test_Python_Mathematics:"

    def test_is_be_able_to_add_two_numbers (self ):
        # ^^ is replaced with "def test_is_able_to_add_two_numbers(self):"
        assert 2 +3 ==5
        assert 2 +0 ==2

    def test_cant_divide_by_zero (self ):
        try :
            2 /0
            assert False ,"Expected an error"
        except ZeroDivisionError :
            pass

# We can also define a class for the describes
# Either when we create the tokeniser and register it
# Or inside the spec file itself, per describe

class DifferentBase (TestCase ):
    def x (self ):
        return 5

class TestInheritance (DifferentBase ):
    def test_has_an_x_equal_to_5 (self ):
        self .assertEqual (self .x (),5 )

# You can even nest describes

class TestNumbers (TestCase ):
    def setUp (self ):
        __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .number1 =1
        self .number2 =2

    def test_has_number1_as_1 (self ):
        self .assertEqual (self .number1 ,1 )

class TestNumbers_TestingNumber3 (TestNumbers ):
    def setUp (self ):
        __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .number3 =3

    def test_has_number1_from_the_lower_level_describe (self ):
        self .assertEqual (self .number1 ,1 )

    def test_also_has_number3 (self ):
        self .assertEqual (self .number3 ,3 )

class TestNumbers_TestingNumber3_LetsChangeANumber (TestNumbers_TestingNumber3 ):
    def setUp (self ):
        __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .number1 =4

    def test_changed_number1_but_kept_others (self ):
        self .assertEqual (self .number1 ,4 )
        self .assertEqual (self .number2 ,2 )
        self .assertEqual (self .number3 ,3 )

TestPythonMathematics .is_noy_spec =True  # type: ignore
TestInheritance .is_noy_spec =True  # type: ignore
TestNumbers .is_noy_spec =True  # type: ignore
TestNumbers_TestingNumber3 .is_noy_spec =True  # type: ignore
TestNumbers_TestingNumber3_LetsChangeANumber .is_noy_spec =True  # type: ignore

TestPythonMathematics .test_cant_divide_by_zero .__testname__ ="can't divide by zero"  # type: ignore



import codecs
import re

import pylama.context
from pylama.config import LOGGER
from pylama.errors import Error
from pylama.lint import Linter as BaseLinter

from noseOfYeti.tokeniser.spec_codec import codec, register

spec_codec = codec()

regexes = {
    "position": re.compile(r"line (\d+), column (\d+)"),
    "encoding": re.compile(r"#\s*coding\s*:\s*spec"),
}


def translate(path: str, contents: str) -> str:
    lines = contents.split("\n")
    pattern = "%(filename)s:%(lnum)s:%(col)s: [%(type)s] %(text)s"

    if contents and regexes["encoding"].match(lines[0]):
        contents = spec_codec.translate(contents)

        if "--- internal spec codec error ---" in lines[-1]:
            found = []

            useful = False
            for line in lines[1:-2]:
                if useful or "SyntaxError:" in line:
                    found.append(line)
                    useful = True

            lnum = "0"
            column = "1"
            for line in found:
                m = regexes["position"].search(line)
                if m:
                    lnum, column = m.groups()
                    break

            msg = "\t".join(found)
            err = Error(linter="pylama", col=int(column), lnum=int(lnum), text=msg, filename=path)
            LOGGER.warning(pattern, err._info)

    return contents


def better_read(filename: str) -> str:
    with codecs.open(filename) as fle:
        return translate(filename, fle.read())


def setup():
    register(transform=True)
    pylama.context.read = better_read


class Linter(BaseLinter):
    def allow(self, path):
        return False

    def run(self, path, **meta):
        return []


setup()



# coding: spec

import shutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

import pytest

try:
    __import__("pylama")
except ImportError:
    pytestmark = pytest.mark.skip("pylama is not current installed")

here = Path(__file__).parent
test_dir = here / "for_formatting_and_pylama"


class TestPylama:
    @pytest.fixture(params=(False, True), ids=["Without black compat", "With black compat"])
    def directory(self, request, tmp_path_factory, monkeypatch):
        directory = tmp_path_factory.mktemp("files")

        monkeypatch.setenv("PYTHONPATH", str(directory))
        monkeypatch.setenv("NOSE_OF_YETI_BLACK_COMPAT", str(request.param).lower())

        with open(directory / "pylama.ini", "w") as fle:
            fle.write(
                dedent(
                    """
            [pylama]
            linters = pylama_noy,mccabe,pycodestyle,pyflakes
            ignore = E225,E202,E211,E231,E226,W292,W291,E251,E122,E501,E701,E227,E305,E128,E228,E203,E261
            """
                )
            )

        if request.param:
            path = Path(__import__("noseOfYeti").__file__).parent / "black" / "noy_black.pth"
            shutil.copy(path, directory)

            with open(directory / "sitecustomize.py ", "w") as fle:
                fle.write(
                    dedent(
                        """
            import site
            import os
            site.addsitedir(os.getcwd())
            """
                    )
                )

        yield directory

    def test_it_lints_when_there_is_black_compat(self, directory):
        shutil.copy(test_dir / "formatted_spec.py", directory / "formatted_spec.py")
        shutil.copy(test_dir / "formatted_normal.py", directory / "formatted_normal.py")

        subprocess.check_call([sys.executable, "-m", "pylama", str(directory)], cwd=directory)

    def test_it_finds_problems(self, directory):
        shutil.copy(test_dir / "failing_pylama_spec.py", directory / "formatted_spec.py")

        p = subprocess.run(
            [sys.executable, "-m", "pylama", str(directory)],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=directory,
            check=False,
        )

        assert p.returncode == 1

        assert (
            p.stdout.decode().strip().replace("\r\n", "\n")
            == dedent(
                """
            formatted_spec.py:3:1 W0611 'sys' imported but unused [pyflakes]
            formatted_spec.py:5:1 E302 expected 2 blank lines, found 1 [pycodestyle]
            formatted_spec.py:7:9 W0612 local variable 'a' is assigned to but never used [pyflakes]
            formatted_spec.py:8:9 E0602 undefined name 'stuff' [pyflakes]
            """
            ).strip()
        )



import fnmatch
import os
import subprocess
import sys
from textwrap import dedent
from unittest import mock

import pytest

from noseOfYeti.tokeniser.spec_codec import Tokeniser, TokeniserCodec

this_dir = os.path.dirname(__file__)
example_dir = os.path.join(this_dir, "..", "example")


def assert_run_subprocess(cmd, expected_output, status=0, **kwargs):
    __tracebackhide__ = True

    # necessary so that tests pass on windows
    env = os.environ.copy()
    env.update(kwargs.get("env", {}))
    kwargs["env"] = env

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs)
    process.wait()

    output = process.stdout.read().decode("utf8").strip().replace("\r\n", "\n")

    print("the process output:")
    print("~" * 80)
    print(output)
    print("~" * 80)

    assert process.returncode == status

    if expected_output is not None:
        pytest.helpers.assert_regex_lines(output, expected_output)

    return output


def assert_glob_lines(got, expect):
    message = [line.strip() for line in got.strip().replace("\r\n", "\n").split("\n")]
    want = [line.strip() for line in expect.strip().split("\n")]

    print("GOT >>" + "=" * 74)
    print()
    print("\n".join(message))
    print()
    print("WANT >>" + "-" * 73)
    print()
    print("\n".join(want))

    count = 1
    while want:
        line = want[0]
        if not message:
            assert False, f"Ran out of lines, stopped at [{count}] '{want[0]}'"

        if message[0] == line or fnmatch.fnmatch(message[0], line):
            count += 1
            want.pop(0)

        message.pop(0)

    if want:
        assert False, f"Didn't match all the lines, stopped at [{count}] '{want[0]}'"


class Test_RegisteringCodec:
    @pytest.fixture()
    def directory(self, tmp_path_factory):
        directory = tmp_path_factory.mktemp("registering_codec")

        with open(directory / "test_code.py", "w") as fle:
            fle.write(
                dedent(
                    """
            # coding: spec

            describe "blah":
                it "should totally work":
                    1 + 1 |should| equal_to(2)

            if hasattr(TestBlah, 'test_should_totally_work'):
                print("test_should_totally_work")
            """
                ).strip()
            )

        with open(directory / "registerer.py", "w") as fle:
            fle.write(
                dedent(
                    """
            import codecs

            try:
                codecs.lookup("spec")
            except LookupError:
                pass
            else:
                assert False, "A spec was already registered!"

            from noseOfYeti.tokeniser import register
            register(transform=True)
            """
                ).strip()
            )

        return directory

    def test_not_registering_codec_leads_to_error(self, directory):
        output = assert_run_subprocess(
            [sys.executable, "-c", "import test_code"],
            expected_output=None,
            cwd=directory,
            env={"NOSE_OF_YETI_BLACK_COMPAT": "false"},
            status=1,
        )
        expected_output = "SyntaxError: [Uu]nknown encoding: spec"
        pytest.helpers.assert_regex_lines(output.split("\n")[-1], expected_output)

    def test_registering_codec_doesnt_lead_to_error(self, directory):
        assert_run_subprocess(
            [sys.executable, "-c", "import registerer; import test_code"],
            expected_output="test_should_totally_work",
            cwd=directory,
            env={"NOSE_OF_YETI_BLACK_COMPAT": "false"},
            status=0,
        )

    def test_can_correctly_translate_file(self):
        with open(os.path.join(example_dir, "test.py")) as fle:
            original = fle.read()
        with open(os.path.join(example_dir, "converted.test.py")) as fle:
            want = fle.read()

        after = TokeniserCodec(Tokeniser()).translate(original)
        pytest.helpers.assert_lines(after, want)

    def test_it_shows_errors(self):
        get_codec = mock.Mock(name="get_codec")
        readline = mock.Mock(name="readline")
        tokeniser = mock.Mock(name="tokeniser")
        tokeniser.translate.side_effect = Exception("NOPE")

        with mock.patch.object(TokeniserCodec, "get_codec", get_codec):
            codec = TokeniserCodec(tokeniser)

            got = codec.dealwith(readline)
            tokeniser.translate.assert_called_once_with(readline, [])

            expect = dedent(
                r'''
                msg = r"""
                Traceback (most recent call last):
                File "*noseOfYeti*tokeniser*spec_codec.py", line *, in dealwith
                Exception: NOPE
                """
                raise Exception(f"--- internal spec codec error --- \n{msg}")
                '''
            )

            assert_glob_lines(got, expect)

            with pytest.raises(Exception) as excinfo:
                exec(got)

            assert excinfo.exconly().startswith("Exception: --- internal spec codec error ---")



# coding: spec

# You ensure the file has the coding: spec comment as the first line as above
# and that nose-of-yeti has registered the spec codec

# The codec will then turn what you have written into python code that can be
# executed.

# The test can then be specified using describes and its

from unittest import TestCase


it "is possible to add numbers":
    assert 1 + 1 == 2

it "is possible to add the number three", three=3:
    # Contrived example of default arguments
    assert 1 + three == 4

describe "Python Mathematics":
    # ^^ is replaced with "class test_Python_Mathematics:"

    it 'is be able to add two numbers':
        # ^^ is replaced with "def test_is_able_to_add_two_numbers(self):"
        assert 2 + 3 == 5
        assert 2 + 0 == 2

    it "can't divide by zero":
        try:
            2 / 0
            assert False, "Expected an error"
        except ZeroDivisionError:
            pass

# We can also define a class for the describes
# Either when we create the tokeniser and register it
# Or inside the spec file itself, per describe

class DifferentBase(TestCase):
    def x(self):
        return 5

describe DifferentBase 'Inheritance':
    it 'has an x equal to 5':
        self.assertEqual(self.x(), 5)

# You can even nest describes

describe TestCase, 'numbers':
    before_each:
        self.number1 = 1
        self.number2 = 2

    it 'has number1 as 1':
        self.assertEqual(self.number1, 1)

    describe 'testing number 3':
        before_each:
            self.number3 = 3

        it 'has number1 from the lower level describe':
            self.assertEqual(self.number1, 1)

        it 'also has number3':
            self.assertEqual(self.number3, 3)

        describe "let's change a number":
            before_each:
                self.number1 = 4

            it 'changed number1 but kept others':
                self.assertEqual(self.number1, 4)
                self.assertEqual(self.number2, 2)
                self.assertEqual(self.number3, 3)



import sys
from textwrap import dedent

import pytest


def assert_example(original, desired):
    tab_original = original.replace("    ", "\t")
    tab_desired = desired.replace("    ", "\t")

    options = {"with_describe_attrs": False}

    for original, desired in ((original, desired), (tab_original, tab_desired)):
        pytest.helpers.assert_conversion(original, desired, tokeniser=options)

        # And with newlines
        original = f"import os\n{dedent(original)}"
        desired = f"import os\n{dedent(desired)}"
        pytest.helpers.assert_conversion(original, desired, tokeniser=options)


class Test_Tokenisor_translation:
    def test_translates_a_describe(self):
        original = 'describe "Something testable"'
        desired = "class TestSomethingTestable :pass"
        assert_example(original, desired)

    def test_translates_an_it(self):
        original = 'it "should do this thing":'
        desired = "def test_should_do_this_thing ()$RET:"
        assert_example(original, desired)

        ## and with async
        original = 'async it "should do this thing":'
        desired = "async def test_should_do_this_thing ()$RET:"
        assert_example(original, desired)

    def test_translates_an_it_with_return_type(self):
        original = 'it "should do this thing" -> None:'
        desired = "def test_should_do_this_thing ()->None :"
        assert_example(original, desired)

        ## and with async
        original = 'async it "should do this thing" -> None:'
        desired = "async def test_should_do_this_thing ()->None :"
        assert_example(original, desired)

    def test_translates_an_it_with_complex_return_type(self):
        original = 'it "should do this thing" -> tp.Generic[Thing, list[str]]:'
        desired = "def test_should_do_this_thing ()->tp .Generic [Thing list [str ]]:"
        assert_example(original, desired)

        ## and with async
        original = 'async it "should do this thing" -> tp.Generic[Thing, list[str]]:'
        desired = "async def test_should_do_this_thing ()->tp .Generic [Thing list [str ]]:"
        assert_example(original, desired)

    def test_translates_an_it_with_return_type_and_args(self):
        original = 'it "should do this thing", one: str -> None:'
        desired = "def test_should_do_this_thing (one :str )->None :"
        assert_example(original, desired)

        ## and with async
        original = 'async it "should do this thing", two: int -> None:'
        desired = "async def test_should_do_this_thing (two :int )->None :"
        assert_example(original, desired)

    def test_translates_an_it_with_complex_return_type_and_args(self):
        original = (
            'it "should do this thing" blah: dict[str, list] -> tp.Generic[Thing, list[str]]:'
        )
        desired = "def test_should_do_this_thing (blah :dict [str ,list ])->tp .Generic [Thing ,list [str ]]:"
        assert_example(original, desired)

        ## and with async
        original = 'async it "should do this thing" item: Item -> tp.Generic[Thing, list[str]]:'
        desired = (
            "async def test_should_do_this_thing (item :Item )->tp .Generic [Thing ,list [str ]]:"
        )
        assert_example(original, desired)

    def test_adds_arguments_to_its_if_declared_on_same_line(self):
        original = 'it "should do this thing", blah, meh:'
        desired = "def test_should_do_this_thing (blah ,meh )$RET:"
        assert_example(original, desired)

    def test_adds_arguments_with_default_string_to_its_if_declared_on_same_line(self):
        original = 'it "should do this thing", blah, meh="hello":'
        desired = 'def test_should_do_this_thing (blah ,meh ="hello")$RET:'
        assert_example(original, desired)

    def test_adds_type_annotations(self):
        original = 'it "should do this thing", blah:str, meh: Thing | Other:'
        desired = "def test_should_do_this_thing (blah :str ,meh :Thing |Other )$RET:"
        assert_example(original, desired)

    def test_allows_comments_after_it(self):
        original = 'it "should do this thing", blah:str, meh: Thing | Other: # a comment'
        desired = "def test_should_do_this_thing (blah :str ,meh :Thing |Other )$RET:# a comment"
        assert_example(original, desired)

        original = 'it "should do this thing":  # a comment'
        desired = "def test_should_do_this_thing ()$RET:# a comment"
        assert_example(original, desired)

    def test_adds_arguments_to_its_if_declared_on_same_line_and_work_with_skipTest(self):
        original = 'it "should do this thing", blah, meh: pass'
        desired = "def test_should_do_this_thing (blah ,meh )$RET:pass"
        assert_example(original, desired)

    def test_complains_about_it_that_isnt_a_block(self):
        with pytest.raises(SyntaxError, match="Found a missing ':' on line 1, column 22"):
            assert_example('it "should be skipped"\n', "")

        with pytest.raises(SyntaxError, match="Found a missing ':' on line 1, column 22"):
            assert_example('it "should be skipped" # blah\n', "")

        # Same tests, but with newlines in front
        with pytest.raises(SyntaxError, match="Found a missing ':' on line 3, column 22"):
            assert_example('import os\n\nit "should be skipped"\n', "")

        original = 'import os\n\nit "should not be skipped":\n'
        desired = "import os\n\ndef test_should_not_be_skipped ()$RET:"
        assert_example(original, desired)

        ## And with async

        with pytest.raises(SyntaxError, match="Found a missing ':' on line 1, column 28"):
            assert_example('async it "should be skipped"\n', "")

        original = 'import os\n\nasync it "should not be skipped":\n'
        desired = "import os\n\nasync def test_should_not_be_skipped ()$RET:"
        assert_example(original, desired)

        # Same tests, but with newlines in front
        with pytest.raises(SyntaxError, match="Found a missing ':' on line 3, column 28"):
            assert_example('import os\n\nasync it "should be skipped"\n', "")

        original = 'import os\n\nasync it "should not be skipped":\n'
        desired = "import os\n\nasync def test_should_not_be_skipped ()$RET:"
        assert_example(original, desired)

    def test_turns_before_each_into_setup(self):
        assert_example("before_each:", "def setUp (self ):")

        # And with async
        assert_example("async before_each:", "async def setUp (self ):")

    def test_indentation_should_work_regardless_of_crazy_groups(self):
        original = """
        describe 'a':
            it 'asdf':
                l = [ True
                    , False
                        , 1
                , 2
                ]

                t = (1
                        , 2
                , 3
            , 4
                    , 5,
                )

                d = {'asdf' : True}

                t2 = (True
                ,    False
        )
            it 'asdf2':
                pass"""

        desired = """
        class TestA :
            def test_asdf (self )$RET:
                l =[True
                ,False
                ,1
                ,2
                ]

                t =(1
                ,2
                ,3
                ,4
                ,5 ,
                )

                d ={'asdf':True }

                t2 =(True
                ,False
                )
            def test_asdf2 (self )$RET:
                pass
        """

        assert_example(original, desired)

    def test_complains_if_describe_after_hanging_it(self):
        original = """
        describe 'thing':
            it 'should be skipped'
            describe 'that':
                pass
        """

        with pytest.raises(SyntaxError, match="Found a missing ':' on line 2, column 26"):
            assert_example(original, "")

    def test_indentation_should_work_for_inline_python_code(self):
        original = """
        describe 'this':
            describe 'that':
                pass

        class SomeMockObject:
            def indented_method()
        """

        desired = """
        class TestThis :pass
        class TestThis_That (TestThis ):
            pass

        class SomeMockObject :
            def indented_method ()
        """

        assert_example(original, desired)

    def test_gives_setups_super_call_when_in_describes_that_know_about_await_if_async(self):
        original = """
        describe "Thing":
            async before_each:
                self.x = 5
        """

        desired = """
        class TestThing :
            async def setUp (self ):
                await __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).async_before_each ();self .x =5
        """

        assert_example(original, desired)

    def test_gives_setups_super_call_when_in_describes(self):
        original = """
        describe "Thing":
            before_each:
                self.x = 5
        """

        desired = """
        class TestThing :
            def setUp (self ):
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();self .x =5
        """

        assert_example(original, desired)

    def test_turns_after_each_into_teardown(self):
        assert_example("after_each:", "def tearDown (self ):")

    def test_gives_teardowns_super_call_that_awaits_when_in_describes_and_async(self):
        original = """
        describe "Thing":
            async after_each:
                self.x = 5
        """

        desired = """
        class TestThing :
            async def tearDown (self ):
                await __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).async_after_each ();self .x =5
        """

        assert_example(original, desired)

    def test_gives_teardowns_super_call_when_in_describes(self):
        original = """
        describe "Thing":
            after_each:
                self.x = 5
        """

        desired = """
        class TestThing :
            def tearDown (self ):
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ();self .x =5
        """

        assert_example(original, desired)

    def test_no_transform_inside_expression(self):
        assert_example("variable = before_each", "variable =before_each ")
        assert_example("variable = after_each", "variable =after_each ")
        assert_example("variable = describe", "variable =describe ")
        assert_example("variable = ignore", "variable =ignore ")
        assert_example("variable = it", "variable =it ")

    def test_sets__testname__on_non_alphanumeric_test_names(self):
        original = """
        it "(root level) should work {well}":
            3 |should| be(4)
        describe "SomeTests":
            it "doesn't get phased by $special characters":
                pass

            describe "NestedDescribe":
                it "asdf $% asdf":
                    1 |should| be(2)
        """

        desired = """
        def test_root_level_should_work_well ()$RET:
            3 |should |be (4 )
        class TestSomeTests :
            def test_doesnt_get_phased_by_special_characters (self )$RET:
                pass

        class TestSomeTests_NestedDescribe (TestSomeTests ):
            def test_asdf_asdf (self )$RET:
                1 |should |be (2 )

        test_root_level_should_work_well .__testname__ ="(root level) should work {well}"  # type: ignore
        TestSomeTests .test_doesnt_get_phased_by_special_characters .__testname__ ="doesn't get phased by $special characters"  # type: ignore
        TestSomeTests_NestedDescribe .test_asdf_asdf .__testname__ ="asdf $% asdf"  # type: ignore
        """

        assert_example(original, desired)

    def test_it_maintains_line_numbers_when_pass_on_another_line(self):
        original = """
        it "is a function with a pass": pass

        it "is a function with a pass on another line":
            pass

        it "is a function with a pass on another line further below":
            #comment or something


            pass

        describe "block with a pass":
            pass

        describe "block with comment and pass":
            # comment or something
            pass

        describe "Nesting and passes": pass
            # comment
            describe "Nested":
                pass

                describe "More Nesting":
                    # comment


                    pass
        """

        desired = """
        def test_is_a_function_with_a_pass ()$RET:pass

        def test_is_a_function_with_a_pass_on_another_line ()$RET:
            pass

        def test_is_a_function_with_a_pass_on_another_line_further_below ()$RET:
            #comment or something


            pass

        class TestBlockWithAPass :
            pass

        class TestBlockWithCommentAndPass :
            # comment or something
            pass

        class TestNestingAndPasses :pass
            # comment
        class TestNestingAndPasses_Nested (TestNestingAndPasses ):
            pass

        class TestNestingAndPasses_Nested_MoreNesting (TestNestingAndPasses_Nested ):
            # comment


            pass
        """

        assert_example(original, desired)

    def test_it_allows_default_arguments_for_its(self):
        original = """
        it "is a test with default arguments", thing=2, other=[3]:
            pass

        describe "group":
            it "has self and default args", blah=None, you=(3, 4,
                5, 5):
                # Test space is respected

                1 |should| be(2)
        """

        desired = """
        def test_is_a_test_with_default_arguments (thing =2 ,other =[3 ])$RET:
            pass

        class TestGroup :
            def test_has_self_and_default_args (self ,blah =None ,you =(3 ,4 ,
            5 ,5 ))$RET:
                # Test space is respected

                1 |should |be (2 )
        """

        assert_example(original, desired)

    def test_can_properly_dedent_after_block_of_just_containers(self):
        original = """
        it "should ensure askers are None or boolean or string":
            for val in (None, False, 'asdf', u'asdf', lambda: 1):
                (lambda : Step(askBeforeAction  = val)) |should_not| throw(Problem)
                (lambda : Step(askDesiredResult = val)) |should_not| throw(Problem)
                (lambda : Step(blockBeforeGet   = val)) |should_not| throw(Problem)

            for val in (1, True):
                (lambda : Step(askBeforeAction  = val)) |should| throw(Problem)
                (lambda : Step(askDesiredResult = val)) |should| throw(Problem)
                (lambda : Step(blockBeforeGet   = val)) |should| throw(Problem)

            3 |should| be(3)
        """

        desired = """
        def test_should_ensure_askers_are_None_or_boolean_or_string ()$RET:
            for val in (None ,False ,'asdf',u'asdf',lambda :1 ):
                (lambda :Step (askBeforeAction =val ))|should_not |throw (Problem )
                (lambda :Step (askDesiredResult =val ))|should_not |throw (Problem )
                (lambda :Step (blockBeforeGet =val ))|should_not |throw (Problem )

            for val in (1 ,True ):
                (lambda :Step (askBeforeAction =val ))|should |throw (Problem )
                (lambda :Step (askDesiredResult =val ))|should |throw (Problem )
                (lambda :Step (blockBeforeGet =val ))|should |throw (Problem )

            3 |should |be (3 )
        """

        assert_example(original, desired)

    def test_it_doesnt_add_semicolon_after_noy_setup_if_not_necessary(self):
        original = """
            describe "block with necessary semicolon":
                before_each:
                    two = 1 + 1

            describe "block with unecessary semiclon":
                before_each:
                    #comment
                    pass

                after_each:

                    pass
        """

        desired = """
        class TestBlockWithNecessarySemicolon :
            def setUp (self ):
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();two =1 +1

        class TestBlockWithUnecessarySemiclon :
            def setUp (self ):
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ()#comment
                pass

            def tearDown (self ):
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ()
                pass
        """

        assert_example(original, desired)

    def test_it_keeps_comments_placed_after_setup_and_teardown_methods(self):
        original = """
            describe "Kls":
                before_each: # Comment one

                    pass

                after_each: # Comment two

                    pass

            describe "Kls2":
                before_each: # Comment three
                    two = 1 + 1

                after_each: # Comment four
                    #comment
                    pass
        """

        desired = """
        class TestKls :
            def setUp (self ):# Comment one
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ()
                pass

            def tearDown (self ):# Comment two
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ()
                pass

        class TestKls2 :
            def setUp (self ):# Comment three
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_before_each ();two =1 +1

            def tearDown (self ):# Comment four
                __import__ ("noseOfYeti").tokeniser .TestSetup (super ()).sync_after_each ()#comment
                pass
        """

        assert_example(original, desired)

    def test_it_doesnt_mess_up_dedent_from_whitespace_in_fstring(self):
        original = """
        def one(
            item: object, want: three, /, _register: four
        ) -> dict | None:
            f"{item} "


        def two(value: object, /) -> dict | None:
            return None
        """

        desired = """
        def one (
        item :object ,want :three ,/,_register :four
        )->dict |None :
            f"{item } "


        def two (value :object ,/)->dict |None :
            return None
        """

        if sys.version_info < (3, 12):
            desired = desired.replace("{item }", "{item}")

        assert_example(original, desired)



import pytest

from noseOfYeti.tokeniser.chooser import TestChooser as Chooser


@pytest.fixture()
def test_chooser():
    return Chooser()


@pytest.fixture()
def Classes():
    class Classes:
        class TestKlsForTest:
            def ignore__test(self):
                pass

            def test_with__test__set(self):
                pass

            test_with__test__set.__test__ = False

            def test_actual(self):
                pass

        class TestIgnoredKls:
            def test_things(self):
                pass

        class TestKlsWithInherited(TestKlsForTest):
            def test_on_subclass(self):
                pass

        class TestKlsParent:
            __only_run_tests_in_children__ = True

            def test_one(self):
                pass

            def test_two(self):
                pass

        class TestKlsChild(TestKlsParent):
            pass

        class TestKlsGrandChild(TestKlsChild):
            pass

    return Classes


class Test_TestChooser:
    def test_it_resets_done_when_told_about_new_module(self, test_chooser):
        assert test_chooser.done == {}

        test_chooser.done["a"] = 3
        assert test_chooser.done == {"a": 3}

        test_chooser.new_module()
        assert test_chooser.done == {}

    def test_already_visited_puts_kls_name_key_in_done_or_returns_True(self, test_chooser):
        assert test_chooser.done == {}
        assert not test_chooser.already_visited("a", "b")
        assert test_chooser.done == {"a.b": True}
        assert test_chooser.already_visited("a", "b")

        assert not test_chooser.already_visited("c", "d")
        assert test_chooser.done == {"a.b": True, "c.d": True}
        assert test_chooser.already_visited("c", "d")


class Test_TestChooser_Consider:
    def test_it_ignores_if_method_starts_with_ignore(self, test_chooser, Classes):
        assert not test_chooser.consider(Classes.TestKlsForTest().ignore__test)

    def test_it_ignores_if_method_has__test__set_to_false(self, test_chooser, Classes):
        assert not test_chooser.consider(Classes.TestKlsForTest().test_with__test__set)

    def test_it_returns_None_if_kls_does_not_have_is_noy_test_set(self, test_chooser, Classes):
        assert test_chooser.consider(Classes.TestKlsForTest().test_actual) is None

    def test_it_ignores_inherited_tests_if_is_noy_test_is_set_on_kls(self, test_chooser, Classes):
        assert test_chooser.consider(Classes.TestKlsWithInherited().test_actual) is None
        Classes.TestKlsWithInherited.is_noy_spec = True
        assert not test_chooser.consider(Classes.TestKlsWithInherited().test_actual)
        assert test_chooser.consider(Classes.TestKlsWithInherited().test_on_subclass)

    def test_it_ignores_functions_already_visited(self, test_chooser, Classes):
        Classes.TestKlsWithInherited.is_noy_spec = True
        assert test_chooser.consider(Classes.TestKlsWithInherited().test_on_subclass)
        assert not test_chooser.consider(Classes.TestKlsWithInherited().test_on_subclass)

    def test_it_ignores_parent_if_specified_to_only_run_tests_in_children(
        self, test_chooser, Classes
    ):
        Classes.TestKlsParent.is_noy_spec = True
        assert not test_chooser.consider(Classes.TestKlsParent().test_one)
        assert not test_chooser.consider(Classes.TestKlsParent().test_two)

    def test_it_runs_parent_tests_in_child_if_specified_in_parent_to_only_run_tests_in_children(
        self, test_chooser, Classes
    ):
        Classes.TestKlsParent.is_noy_spec = True
        Classes.TestKlsChild.is_noy_spec = True
        assert test_chooser.consider(Classes.TestKlsChild().test_one)
        assert test_chooser.consider(Classes.TestKlsChild().test_two)

    def test_it_doesnt_run_grandparent_tests_if_specified_in_grandparent_to_only_run_tests_in_children(
        self,
        test_chooser,
        Classes,
    ):
        Classes.TestKlsParent.is_noy_spec = True
        Classes.TestKlsChild.is_noy_spec = True
        Classes.TestKlsGrandChild.is_noy_spec = True
        assert not test_chooser.consider(Classes.TestKlsGrandChild().test_one)
        assert not test_chooser.consider(Classes.TestKlsGrandChild().test_two)



import shutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

here = Path(__file__).parent


class TestPyTest:
    def test_it_collects_tests_correctly(self, tmp_path_factory, monkeypatch):

        directory = tmp_path_factory.mktemp("files")

        monkeypatch.setenv("NOSE_OF_YETI_BLACK_COMPAT", "false")

        shutil.copytree(here / "for_pytest_plugin", directory / "tests")

        want = set(
            dedent(
                """
            tests/test_one.py::test_one
            tests/test_one.py::TestTwo::test_three
            tests/test_one.py::TestTwo::test_four
            tests/test_one.py::TestTwo_Five::test_six
            tests/test_one.py::TestTwo_Seven_Ten::test_eight
            tests/test_one.py::TestTwo_Seven_Ten::test_nine
            tests/test_one.py::TestTwo_Seven_Eleven::test_eight
            tests/test_one.py::TestTwo_Seven_Eleven::test_nine
            tests/test_one.py::TestTwelve::test_thirteen
            tests/test_two.py::TestOne::test_two
            tests/test_two.py::TestOne::test_three
            tests/test_two.py::TestOne::test_four
            tests/test_two.py::TestOne::test_five
            tests/test_two.py::TestOne_Six::test_seven
            tests/test_two.py::TestOne_Eight::test_nine
            tests/test_two.py::TestOne_Ten::test_eleven
            """
            )
            .strip()
            .split("\n")
        )

        with subprocess.Popen(
            [
                sys.executable,
                "-m",
                "pytest",
                str(directory / "tests" / "test_one.py"),
                str(directory / "tests" / "test_two.py"),
                "-v",
                "-o",
                "console_output_style=short",
            ],
            cwd=directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as process:
            try:
                process.wait(timeout=5)
            except:
                process.kill()
                raise

            out = process.stdout.read().decode()

            got = set()
            for line in out.split("\n"):
                if "PASSED" in line:
                    got.add(line.split(" ", 1)[0])

            print(out)

            print("WANT:")
            for line in sorted(want):
                print("  ", line)

            print()
            print("GOT:")
            for line in sorted(got):
                print("  ", line)

            assert "16 passed" in out
            assert got == want



import re
from tokenize import COMMENT, NAME

regexes = {
    "joins": re.compile(r"[- /]"),
    "repeated_underscore": re.compile(r"_{2,}"),
    "invalid_variable_name_start": re.compile(r"^[^a-zA-Z_]"),
    "invalid_variable_name_characters": re.compile(r"[^0-9a-zA-Z_]"),
}


def acceptable(name, capitalize=False):
    """Convert a string into something that can be used as a valid python variable name"""
    # Convert space and dashes into underscores
    name = regexes["joins"].sub("_", name)

    # Remove invalid characters
    name = regexes["invalid_variable_name_characters"].sub("", name)

    # Remove leading characters until we find a letter or underscore
    name = regexes["invalid_variable_name_start"].sub("", name)

    # Clean up irregularities in underscores.
    name = regexes["repeated_underscore"].sub("_", name.strip("_"))

    if capitalize:
        # We don't use python's built in capitalize method here because it
        # turns all upper chars into lower chars if not at the start of
        # the string and we only want to change the first character.
        name_parts = []
        for word in name.split("_"):
            name_parts.append(word[0].upper())
            if len(word) > 1:
                name_parts.append(word[1:])
        name = "".join(name_parts)

    return name


class TokenDetails:
    """Container for current token"""

    def __init__(self, tokenum=None, value=None, srow=0, scol=0):
        self.set(tokenum, value, srow, scol)

    def set(self, tokenum, value, srow, scol):
        self.srow = srow
        self.scol = scol
        self.value = value
        self.tokenum = tokenum

    def transfer(self, details):
        details.set(*self.values())

    def values(self):
        return self.tokenum, self.value, self.srow, self.scol


class Single:
    """Container for a single block (i.e. it or ignore block)"""

    def __init__(self, srow, scol, group, typ=None, indent=0):
        self.typ = typ
        self.group = group
        self.indent = indent

        self.args = []
        self.comments = []
        self.return_type = None

        self._name = None
        self.english = None
        self.skipped = False

        self.started_arg = False
        self.starting_arg = False

        if not self.group.root:
            self.args.append((NAME, "self", srow, scol))
        else:
            self.args.append((None, None, srow, scol))

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = acceptable(value)
        self.english = value

    @property
    def python_name(self):
        if self.typ == "it":
            return f"test_{self.name}"
        else:
            return f"ignore__{self.name}"

    @property
    def identifier(self):
        if self.group.root:
            return self.python_name
        else:
            return f"{self.group.kls_name}.{self.python_name}"

    def add_to_arg(self, tokenum, value, srow, scol):
        if value == "," and len(self.args) == 1 and self.args[0][0] is None:
            return

        if value == "->" and self.return_type is None:
            self.return_type = [(tokenum, value, srow, scol)]
            return

        if self.comments or tokenum is COMMENT:
            self.comments.append((tokenum, value))
        elif self.return_type is not None:
            self.return_type.append((tokenum, value, srow, scol))
        else:
            self.args.append((tokenum, value, srow, scol))


class Group:
    """Container for describe blocks"""

    def __init__(self, name=None, root=False, parent=None, level=0, typ=None):
        self.kls = None
        self.typ = typ
        self.name = name
        self.root = root
        self.empty = True
        self.level = level
        self.parent = parent
        self.singles = []
        self.has_after_each = False
        self.async_after_each = False
        self.has_before_each = False
        self.async_before_each = False

        # Default whether this group is starting anything
        self.starting_single = False
        if root:
            # Root technically isn't a group, so it doesn't have a signature to start
            self.starting_group = False
        else:
            # Group is created before we have all information
            # Hence it's signature is being created
            self.starting_group = True

    def __repr__(self):
        if self.root:
            return "<Root Group>"
        else:
            return f"<Group {self.kls_name}:{self.super_kls}({self.parent})>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.english = value
        if value:
            self._name = acceptable(value, True)

    @property
    def starting_signature(self):
        """Determine if this group is starting itself or anything belonging to it"""
        return self.starting_group or self.starting_single

    @property
    def kls_name(self):
        """Determine python name for group"""
        # Determine kls for group
        if not self.parent or not self.parent.name:
            return f"Test{self.name}"
        else:
            use = self.parent.kls_name
            if use.startswith("Test"):
                use = use[4:]

            return f"Test{use}_{self.name}"

    @property
    def super_kls(self):
        """
        Determine what kls this group inherits from
        If default kls should be used, then None is returned
        """
        if not self.kls and self.parent and self.parent.name:
            return self.parent.kls_name
        return self.kls

    def start_group(self, scol, typ):
        """Start a new group"""
        return Group(parent=self, level=scol, typ=typ)

    def start_single(self, typ, srow, scol):
        """Start a new single"""
        self.starting_single = True
        single = self.single = Single(srow, scol, self, typ=typ, indent=(scol - self.level))
        self.singles.append(single)
        return single

    def finish_signature(self):
        """Tell group it isn't starting anything anymore"""
        self.starting_group = False
        self.starting_single = False

    def modify_kls(self, name):
        """Add a part to what will end up being the kls' superclass"""
        if self.kls is None:
            self.kls = name
        else:
            self.kls += name



import pytest


class Examples:
    small_example = [
        """
    describe "This":
        describe "That":

            describe "Meh":pass
        describe "Blah":pass
    describe "Another":pass """,
        """
    class TestThis :pass
    class TestThis_That (TestThis ):pass
    class TestThis_That_Meh (TestThis_That ):pass
    class TestThis_Blah (TestThis ):pass
    class TestAnother :pass
    """,
    ]

    ###   SMALL EXAMPLE (WITH PATH FOR BW COMPAT)

    small_example_with_pass = [
        """
    describe "This":pass
        describe "That":pass
            describe "Meh":pass
        describe "Blah":pass
    describe "Another":pass """,
        """
    class TestThis :pass
    class TestThis_That (TestThis ):pass
    class TestThis_That_Meh (TestThis_That ):pass
    class TestThis_Blah (TestThis ):pass
    class TestAnother :pass
    """,
    ]

    big_example = [
        """
    describe "This":
        it 'should':
            if x:
                pass
            else:
                x += 9
        describe "That":
            describe "Meh":
                it 'should':
                    if y:
                        pass
                    else:
                        pass
        describe "Blah":pass
    describe "Another":
        it 'should':
            if z:
                if u:
                    print "hello \
                        there"
                else:
                    print "no"
            else:
                pass
    """,
        """
    class TestThis :
        def test_should (self )$RET:
            if x :
                pass
            else :
                x +=9
    class TestThis_That (TestThis ):pass
    class TestThis_That_Meh (TestThis_That ):
        def test_should (self )$RET:
            if y :
                pass
            else :
                pass
    class TestThis_Blah (TestThis ):pass
    class TestAnother :
        def test_should (self )$RET:
            if z :
                if u :
                    print "hello \
                        there"
                else :
                    print "no"
            else :
                pass
    """,
    ]


def assert_example(example, convert_to_tabs=False):
    __tracebackhide__ = True
    pytest.helpers.assert_example(
        example, convert_to_tabs=convert_to_tabs, with_describe_attrs=False
    )


class Test_Tokeniser_Nesting:
    def test_works_with_space(self):
        assert_example(Examples.small_example)

    def test_works_with_tabs(self):
        assert_example(Examples.small_example, convert_to_tabs=True)

    def test_works_with_space_and_inline_pass(self):
        assert_example(Examples.small_example_with_pass)

    def test_works_with_tabs_and_inline_pass(self):
        assert_example(Examples.small_example_with_pass, convert_to_tabs=True)

    def test_keeps_good_indentation_in_body_with_spaces(self):
        assert_example(Examples.big_example)

    def test_keeps_good_indentation_in_body_with_tabs(self):
        assert_example(Examples.big_example, convert_to_tabs=True)

    def test_names_nested_describes_with_part_of_parents_name(self):
        original = 'describe "a":\n\tdescribe "b":'
        desired = "class TestA :pass\nclass TestA_B (TestA ):"
        assert_example([original, desired])



import os
from tokenize import generate_tokens

from noseOfYeti.tokeniser.tokens import Tokens
from noseOfYeti.tokeniser.tracker import Tracker

WITH_IT_RETURN_TYPE_ENV_NAME = "NOSE_OF_YETI_IT_RETURN_TYPE"


class Tokeniser:
    """Endpoint for tokenising a file"""

    def __init__(self, with_describe_attrs=True, with_it_return_type=None):
        if with_it_return_type is None:
            self.with_it_return_type = os.environ.get(
                WITH_IT_RETURN_TYPE_ENV_NAME, ""
            ).lower() not in ("", "false", "0")
        else:
            self.with_it_return_type = with_it_return_type
        self.with_describe_attrs = with_describe_attrs

    def translate(self, readline, result=None, no_imports=None):
        # Tracker to keep track of information as the file is processed
        self.tokens = Tokens()
        self.tracker = Tracker(result, self.tokens, with_it_return_type=self.with_it_return_type)

        try:
            # Looking at all the tokens
            with self.tracker.add_phase() as tracker:
                for tokenum, value, (srow, scol), _, _ in generate_tokens(readline):
                    tracker.next_token(tokenum, value, srow, scol)
        finally:
            # Complain about mismatched brackets
            self.tracker.raise_about_open_containers()

        # Add attributes to our Describes so that the plugin can handle some nesting issues
        # Where we have tests in upper level describes being run in lower level describes
        if self.with_describe_attrs:
            self.tracker.make_describe_attrs()

        # Add lines to bottom of file to add __testname__ attributes
        self.tracker.make_method_names()

        # Return translated list of tokens
        return self.tracker.result



# coding: spec

import pytest


@pytest.fixture()
def register():
    return 2


describe "one":
    it "two", register:
        assert register == 2

    it "three", register:
        assert register == 2

    it "four", register:
        assert register == 2

    it "five", register:
        assert register == 2

    describe "six":

        @pytest.fixture()
        def collector(self):
            return 3

        async it "seven", register, collector:
            assert register == 2
            assert collector == 3

    describe "eight":

        @pytest.fixture()
        def collector(self):
            return 4

        async it "nine", register, collector:
            assert register == 2
            assert collector == 4

    describe "ten":

        @pytest.fixture()
        def collector(self):
            return 5

        @pytest.fixture()
        def superman(self, collector):
            return collector * 2

        @pytest.fixture()
        def register(self):
            return 20

        it "eleven", superman, register, collector:
            assert collector == 5
            assert superman == 10
            assert register == 20



# coding: spec


def awesome(a: str) -> bool:
    return True


def hi(twos: int, word: str, b: bool):
    pass


describe "TestThing":
    async it "totes works yo", one: int, three: str:
        assert awesome(2)  # type: ignore


it "is great":
    assert False, "or is it ?"


def with_other_things():
    hi(22222, "asdfasdf", True)



import shutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

import pytest

from noseOfYeti.plugins.black_compat import is_supported_black_version

try:
    import black

    if not is_supported_black_version():
        pytestmark = pytest.mark.skip(
            f"Currently installed black is an unsupported version {black.__version__}"
        )
except ImportError:
    pytestmark = pytest.mark.skip("black is not current installed")

here = Path(__file__).parent
test_dir = here / "for_formatting_and_pylama"


class TestBlackFormatting:
    def test_it_formats(self, tmp_path_factory, monkeypatch):

        from black.cache import CACHE_DIR

        if CACHE_DIR.exists():
            for fle in CACHE_DIR.iterdir():
                if "noy" in fle.name:
                    fle.unlink()

        directory = tmp_path_factory.mktemp("files")

        monkeypatch.setenv("PYTHONPATH", str(directory))
        monkeypatch.setenv("NOSE_OF_YETI_BLACK_COMPAT", "true")

        with open(directory / "sitecustomize.py", "w") as fle:
            fle.write(
                dedent(
                    """
            import site
            import os
            site.addsitedir(os.getcwd())
            """
                )
            )

        shutil.copy(test_dir / "unformatted_spec.py", directory / "unformatted_spec.py")
        shutil.copy(test_dir / "unformatted_normal.py", directory / "unformatted_normal.py")

        path = Path(__import__("noseOfYeti").__file__).parent / "black" / "noy_black.pth"
        shutil.copy(path, directory)

        subprocess.check_call([sys.executable, "-m", "black", str(directory)], cwd=directory)

        with open(directory / "unformatted_spec.py") as spec_result, open(
            test_dir / "formatted_spec.py"
        ) as spec_want:
            assert spec_result.read() == spec_want.read()

        with open(directory / "unformatted_normal.py") as normal_result, open(
            test_dir / "formatted_normal.py"
        ) as normal_want:
            assert normal_result.read() == normal_want.read()



# coding: spec

def awesome(a: str)-> bool:
    return  True

def hi(    twos: int, word: str, b: bool)    :
    pass
describe  "TestThing" :
    async  it "totes works yo", one:   int, three: str:
        assert   awesome(2)   # type: ignore


it  "is great":
    assert False,  "or is it ?"

def with_other_things(   ):
    hi (  22222, "asdfasdf",True)



from nose.plugins import Plugin

from noseOfYeti.tokeniser.chooser import TestChooser
from noseOfYeti.tokeniser.spec_codec import register


class Plugin(Plugin):
    name = "noseOfYeti"

    def __init__(self, *args, **kwargs):
        self.test_chooser = TestChooser()
        super(Plugin, self).__init__(*args, **kwargs)

    def options(self, parser, env={}):
        super(Plugin, self).options(parser, env)

        parser.add_option(
            "--with-noy",
            default=False,
            action="store_true",
            dest="enabled",
            help="Enable nose of yeti",
        )

    def wantModule(self, mod):
        self.test_chooser.new_module()

    def wantMethod(self, method):
        return self.test_chooser.consider(method)

    def configure(self, options, conf):
        super(Plugin, self).configure(options, conf)

        if options.enabled:
            self.done = {}
            self.enabled = True
            register(transform=True)



import re

from pyls import hookimpl

from noseOfYeti.tokeniser.spec_codec import codec

spec_codec = codec()

regexes = {
    "encoding": re.compile(r"#\s*coding\s*:\s*spec"),
    "first_whitespace": re.compile(r"^(\s*)"),
}


@hookimpl(hookwrapper=True)
def pyls_initialize(config, workspace):
    spec_codec.register(transform=True)
    yield


@hookimpl(hookwrapper=True)
def pyls_document_did_open(config, workspace, document):
    contents = document._source
    lines = contents.split("\n")

    if contents and regexes["encoding"].match(lines[0]):
        translated = spec_codec.translate(contents)
        translated = translated.split("\n")[: len(lines)]

        replacement = []
        for orig, new in zip(lines, translated):
            if new.startswith("class"):
                # We still need the root classes to be at the start of the document
                # So that their parent classes exist at that scope
                replacement.append(new)
            else:
                # Everything else however needs to be at their original indentation
                # So that pyls doesn't get confused by columns
                or_space = regexes["first_whitespace"].search(orig).groups()[0]
                tr_space = regexes["first_whitespace"].search(new).groups()[0]
                replacement.append(f"{or_space}{new[len(tr_space):]}")

        document._source = "\n".join(replacement)

    yield
